%{
/***************************************************************************************
 * This is the first version of my ocl parser
 * Some of the production rules defined in the specification are syntactically
 * ambiguouts, so disambiguating rules have also been defined in the spicification.
 * But those rules may be based on the UML model to which the OCL expression
 * is attached, thus the UML model must be available when an OCL expression is
 * parsed using those production rules.
 * At the time being, I couldn't figure out how to access the UML model info.
 * So I will change the production rules quite a bit, which means at the end
 * of this first parse, an instance of the abstract syntax could not be build.
 * My final option may be the grammar given in OCL1.4 specification. But I won't
 * chose it as long as I have another choice.
 * 
 ***************************************************************************************
 *
 * Notes:
 * 1.The concrete syntax in the OCL2.0 specification does not define
 *   IntegerLiteralExp, RealLiteralExp, StringLiteralExp and BooleanLiteralExp
 *   for internationalization, I make my own decisions here, but they may be
 *   modified later.
 * 2.In the concrete syntax, EnumLiteralExp is defined as:
 *      EnumLiteralExp ::= name '::' <String>
 *   with the disambiguating rule: the specified name must indeed reference
 *   an enumeration. Or, it will have a reduce/reduce conflict with the
 *   production rule for name.
 *   But it is not easy to checked. So, I postpone this question by modifying
 *   the rule as below:
 *       EnumLiteralExp ::= '#' name '::' <String>
 * 3.CollectionRange ::= OclExpression '..' OclExpression
 *   Since an OclExpression could be a real number, to make "1..3" parsed correctly,
 *   I decide that "1." is an illegal form of a real number.
 * 4.According to the abstract syntax, the iterator for LoopExp is mandatary.
 *   So, instead of define iterator as:
 *       iterator::= VariableDeclaration?
 *   we could have an iterator or a default iterator, but even by doing that,
 *   the '|' is mandatary, or we will end up with a shift/reduce conflict
 *   I won't use a default_iterator here.
 * 5.In a VarDeclaration, if an initExpression is given, the type must be also given.
 * 6.A referredOperation can be a relational operation, a predefined collection
 *   operation or a operation "mdlProperty" defined in the UML model by the user
 *   "Implies" is added to the "RELOP" list; "not" operation is also added
 *   Furthermore, operationCall, attributeCall and navigationCall will share the keyword
 *   mdlProperty to resolve conflicts. An attributeCall and a navigationCall will be the
 *   same at this stage without knowing the UML model.
 * 7.To parse correctly, the source of the OperationCallExp must be specified
 * 9.The unary minus is not handled at the time being
 *10.OclOperationWithStateArgument operation is not handled at the time being
 *11.The syntax of LetExp is not the same as that in the specification
 *   the semi-colon is specified in the concrete syntax but not in the examples
 *   given in the specification, but i decide to keep the semi-colon
 *12.The precedence of '^' is not given in the concrete syntax, I decide to give it
 *   the same precedence as dot and arrow.
 *   The OclMessageExp need to be studied further.
 *13.TRUE and FALSE will be defined as boolean value on the one hand;
 *   on the other hand, they will be also return from lex as tokens SYM_TRUE and SYM_FALSE
 *14.Token REJECT will be noted as SYM_REJECT, since REJECT is a lex keyword
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include "util.h"
#include "absyn.h"
#include "creat_python.h"

extern FILE *yyin, *yyout;

int lineno = 1;


%}



%union {
  int ival;
  double dval;
  string sval;
  bool bval;
  A_oclExpression oclExpression;
  A_propertyCallExp propertyCallExp;
  A_variableExp variableExp;
  A_literalExp literalExp;
  A_letExp letExp;
  A_oclMessageExp oclMessageExp;
  A_ifExp ifExp;
  A_name pathName;
  A_enumLiteralExp enumLiteralExp;
  A_collectionLiteralExp collectionLiteralExp;
  A_tupleLiteralExp tupleLiteralExp;
  A_numericLiteralExp numericLiteralExp;
  A_stringLiteralExp stringLiteralExp;
  A_booleanLiteralExp booleanLiteralExp;
  A_collectionTypeIdentifier collectionType;
  A_collectionItem collectionItem;
  A_collectionRange collectionRange;
  A_integerLiteralExp integerLiteralExp;
  A_realLiteralExp realLiteralExp;
  A_tupleLiteralExpParts tupleLiteralExpParts;
  A_variableDeclaration variableDeclaration;
  A_modelPropertyCallExp modelPropertyCallExp;
  A_loopExp loopExp;
  A_iteratorExp iteratorExp;
  A_iterateExp iterateExp;
  A_collectionOp collectionOp;
  A_operationCallExp operationCallExp;
  A_attrOrNavCallExp attrOrNavCallExp;
  A_variables vars;
  A_oclMessageArg oclMessageArg;
  A_arguments args;
}

%token <ival> INTEGER
%token <dval> REAL
%token <sval> QSTRING NAMESTRING

%token SYM_TRUE SYM_FALSE

%token SELECT COLLECT SYM_REJECT FORALL EXISTS ISUNIQUE SIZE ISEMPTY
%token NOTEMPTY COUNT UNION DIFFERENCE NOT

%token INV PRE POST DEF
%token SET BAG SEQUENCE TUPLE
%token DCOLON DDOT IN ITERATE ATPRE LET
%token IF THEN ELSE ENDIF

%type <oclExpression> OclExpression individualItem body initExpression argument
%type <oclExpression> LetExpSub scope condition thenExpression elseExpression
%type <propertyCallExp> PropertyCallExp
%type <variableExp> VariableExp
%type <literalExp> LiteralExp
%type <letExp> LetExp
%type <oclMessageExp> OclMessageExp
%type <ifExp> IfExp
%type <pathName> name type
%type <enumLiteralExp> EnumLiteralExp
%type <collectionLiteralExp> CollectionLiteralExp
%type <tupleLiteralExp> TupleLiteralExp
%type <numericLiteralExp> NumericLiteralExp
%type <stringLiteralExp> StringLiteralExp
%type <booleanLiteralExp> BooleanLiteralExp
%type <collectionType> CollectionTypeIdentifier
%type <collectionItem> CollectionItem CollectionItems CollectionItemsPart
%type <collectionRange> CollectionRange
%type <integerLiteralExp> IntegerLiteralExp;
%type <realLiteralExp> RealLiteralExp;
%type <stringLiteralExp> StringLiteralExp;
%type <booleanLiteralExp> BooleanLiteralExp;
%type <tupleLiteralExpParts> TupleLiteralExpParts
%type <variableDeclaration> VariableDeclaration TupleLiteralExpPart iterator result
%type <variableDeclaration> variableDeclarationInLetExp variable
%type <modelPropertyCallExp> ModelPropertyCallExp
%type <loopExp> LoopExp
%type <iteratorExp> IteratorExp
%type <iterateExp> IterateExp
%type <collectionOp> colop
%type <sval> varName
%type <operationCallExp> OperationCallExp
%type <attrOrNavCallExp> AttrOrNavCallExp
%type <bval> timeExpr
%type <args> arguments argumentsOrEmpty
%type <sval> mdlProperty
%type <vars> variables
%type <oclMessageArg> OclMsgArgOrNull OclMessageArgs OclMessageArg


%left IMPLIES
%left AND OR XOR
%left EQ NOTEQ
%left LT GT LTEQ GTEQ
%left PLUS MINUS
%left TIMES DIVIDES
%left NOT
%left DOT ARROW '^'

%%
start: OclConstraints  { fprintf(yyout, "start\n"); /* setup the initial env of an expression */ }
  ;

OclConstraints: /* empty */
  | OclConstraints OclConstraint  { fprintf(yyout, "OclExpressions\nAt line: %d\n\n",lineno); }
  ;

OclConstraint: constraintType ':' OclExpression { traverse($3, 0); }
  ;

constraintType: INV
  | PRE
  | POST
  | DEF
  ;

OclExpression:  PropertyCallExp { $$=A_OclExpression($1,PROPERTYCALLEXP); fprintf(yyout, "PropertyCallExp\n"); }
  | VariableExp { $$=A_OclExpression($1,VARIABLEEXP); fprintf(yyout, "VariableExp\n"); }
  | LiteralExp  { $$=A_OclExpression($1,LITERALEXP); fprintf(yyout, "LiteralExp\n"); }
  | LetExp  { $$=A_OclExpression($1,LETEXP); fprintf(yyout, "LetExp\n"); }
  | OclMessageExp { $$=A_OclExpression($1,OCLMESSAGEEXP); fprintf(yyout, "OclMessageExp\n"); }
  | IfExp { $$=A_OclExpression($1,IFEXP); fprintf(yyout, "IfExp\n"); }
  | '(' OclExpression ')' { $$=$2; fprintf(yyout, "Parenthesized OclExpression\n"); }
  ;


  /* define VariableExp */
VariableExp:  name  { $$=A_VariableExp($1);
                      fprintf(yyout, "name\n");
                      /* Disambiguating rule
                       * name must be a name of a visible VaribleDeclaration in the current env */
                    }
  ;

name: NAMESTRING  { $$ = A_Name($1,NULL); fprintf(yyout, "namestring\n"); }
  | NAMESTRING DCOLON name  { $$ = A_Name($1,$3); fprintf(yyout, "name::namestring\n"); }
  ;


  /* define LiteralExp */
LiteralExp: EnumLiteralExp   { $$=A_LiteralExp($1,ENUMLITERALEXP); fprintf(yyout, "EnumLiteralExp\n"); }
  | CollectionLiteralExp  { $$=A_LiteralExp($1,COLLECTIONLITERALEXP); fprintf(yyout, "CollectionLiteralExp\n"); }
  | TupleLiteralExp  { $$=A_LiteralExp($1,TUPLELITERALEXP); fprintf(yyout, "TupleLiteralExp\n"); }
  | NumericLiteralExp  { $$=A_LiteralExp($1,NUMERICLITERALEXP); fprintf(yyout, "NumericLiteralExp\n"); }
  | StringLiteralExp   { $$=A_LiteralExp($1,STRINGLITERALEXP); fprintf(yyout, "StringLiteralExp\n"); }
  | BooleanLiteralExp  { $$=A_LiteralExp($1,BOOLEANLITERALEXP); fprintf(yyout, "BooleanLiteralExp\n"); }
  ;


EnumLiteralExp: '#' name DCOLON NAMESTRING { $$=A_EnumLiteralExp($2,$4); fprintf(yyout, "#name::namestring\n"); }
  ;


  /* define CollectionLiteralExp */
CollectionLiteralExp: CollectionTypeIdentifier '{' CollectionItemsPart '}'
  { $$=A_CollectionLiteralExp($1,$3); fprintf(yyout, "CollectionTypeIdentifier '{' CollectionItemsPart '}'\n"); }
  ;

CollectionTypeIdentifier: SET  { $$=A_SET;fprintf(yyout, "CollectionTypeIdentifier:Set\n"); }
  | BAG  { $$=A_BAG;fprintf(yyout, "CollectionTypeIdentifier:Bag\n"); }
  | SEQUENCE  { $$=A_SEQUENCE;fprintf(yyout, "CollectionTypeIdentifier:Sequence\n"); }
  ;

CollectionItemsPart:  /* empty */  { $$ = NULL; fprintf(yyout, "empty collection\n"); }
  | CollectionItems   { $$ = $1; }
  ;

CollectionItems:  CollectionItem  { $$=$1; }
  | CollectionItem ',' CollectionItems  { $$=A_CollectionItem_link($1,$3);
                                          fprintf(yyout, "CollectionItems , CollectionItem\n"); }
  ;

CollectionItem: CollectionRange  { $$=A_CollectionItem($1,COLLECTIONRANGE); fprintf(yyout, "CollectionRange\n"); }
  | individualItem  { $$=A_CollectionItem($1,OCLEXPRESSION); fprintf(yyout, "IndividualItem\n"); }
  ;

CollectionRange:  OclExpression DDOT OclExpression { $$=A_CollectionRange($1,$3); fprintf(yyout, "OclExp..OclExp\n"); }
  ;

individualItem: OclExpression { $$=$1; fprintf(yyout, "OclExp\n"); }
  ;


  /* define NumericLiteralExp */
NumericLiteralExp:  IntegerLiteralExp  { $$=A_NumericLiteralExp($1,INTEGERLITERALEXP);
                                         fprintf(yyout, "IntegerLiteralExp\n"); }
  | RealLiteralExp  { $$=A_NumericLiteralExp($1,REALLITERALEXP); fprintf(yyout, "RealLiteralExp\n"); }
  ;

IntegerLiteralExp: INTEGER  { $$=A_IntegerLiteralExp($1); fprintf(yyout, "Integer\n"); }
  ;

RealLiteralExp: REAL  { $$=A_RealLiteralExp($1); fprintf(yyout, "RealExp\n"); }
  ;

  /* define StringLiteralExp */
StringLiteralExp: QSTRING  { $$=A_StringLiteralExp($1); fprintf(yyout, "QStringlExp\n"); }
  ;

  /* define BooleanLiteralExp */
BooleanLiteralExp: SYM_TRUE  { $$=A_BooleanLiteralExp(TRUE); fprintf(yyout, "true\n"); }
  | SYM_FALSE  { $$=A_BooleanLiteralExp(FALSE); fprintf(yyout, "false\n"); }
  ;


  /* define TupleLiteralExp */
TupleLiteralExp: TUPLE '{' TupleLiteralExpParts '}' { $$=A_TupleLiteralExp($3);
                                                      fprintf(yyout, "TUPLE '{' non-empty TupleLiteralExpParts '}'\n"); }
  | TUPLE '{' '}' { $$=A_TupleLiteralExp(NULL); }
  ;

TupleLiteralExpParts: TupleLiteralExpPart  { $$=A_TupleLiteralExpParts($1,NULL); }
  | TupleLiteralExpPart ',' TupleLiteralExpParts  { $$=A_TupleLiteralExpParts($1,$3); }
  ;

TupleLiteralExpPart:  VariableDeclaration  { $$=$1; fprintf(yyout, "TupleLiteralExpPart\n"); }
  ;


  /* define PropertyCallExp */
PropertyCallExp:  ModelPropertyCallExp  { $$=A_PropertyCallExp($1,MODELPROPERTYCALLEXP);
                                          fprintf(yyout, "ModelPropertyCallExp\n"); }
  | LoopExp  { $$=A_PropertyCallExp($1,LOOPEXP); fprintf(yyout, "LoopExp\n"); }
  ;

  /* define LoopExp */
LoopExp:  IteratorExp { $$=A_LoopExp($1,ITERATOREXP); fprintf(yyout, "IteratorExp\n"); }
  | IterateExp  { $$=A_LoopExp($1,ITERATEEXP); fprintf(yyout, "IterateExp\n"); }
  ;

  /* define IteratorExp */
IteratorExp:  OclExpression ARROW colop '(' iterator '|' body ')'
    { $$=A_IteratorExp($1,$3,$5,$7); }
  ;

  /* define IterateExp */
IterateExp: OclExpression ARROW ITERATE '(' iterator ';' result '|' body ')'
    { $$=A_IterateExp($1,$5,$7,$9); }
  ;



colop:  SELECT  { $$=A_SELECT; fprintf(yyout, "select\n"); }
  | COLLECT { $$=A_COLLECT; fprintf(yyout, "collect\n"); }
  | SYM_REJECT  { $$=A_REJECT; fprintf(yyout, "reject\n"); }
  | FORALL  { $$=A_FORALL; fprintf(yyout, "forAll\n"); }
  | EXISTS  { $$=A_EXISTS; fprintf(yyout, "exists\n"); }
  | ISUNIQUE  { $$=A_ISUNIQUE; fprintf(yyout, "isUnique\n"); }
  | SIZE    { $$=A_SIZE; fprintf(yyout, "size\n"); }
  | ISEMPTY { $$=A_ISEMPTY; fprintf(yyout, "isEmpty\n"); }
  | NOTEMPTY  { $$=A_NOTEMPTY; fprintf(yyout, "notEmpty\n"); }
  | COUNT   { $$=A_COUNT; fprintf(yyout, "count\n"); }
  | UNION   { $$=A_UNION; fprintf(yyout, "union\n"); }
  | DIFFERENCE  { $$=A_DIFFERENCE; fprintf(yyout, "difference\n"); }
  ;

iterator: /* empty */  { $$=NULL; fprintf(yyout, "default iterator\n"); }
  | VariableDeclaration  { $$=$1; fprintf(yyout, "iterator is a VariableDeclaration\n"); }
  ;

result: VariableDeclaration  { $$=$1; fprintf(yyout, "result is a VariableDeclaration\n"); }
  ;

  /* define VariableDeclaration */
VariableDeclaration:  varName ':' type EQ initExpression
    { $$=A_VariableDeclaration($1,$3,$5);
      fprintf(yyout, "VariableDeclaration with type and initExpression\n"); }
  | varName ':' type  { $$=A_VariableDeclaration($1,$3,NULL);
                        fprintf(yyout, "VariableDeclaration with type but not initExpression\n"); }
  | varName  { $$=A_VariableDeclaration($1,NULL,NULL);
               fprintf(yyout, "VariableDeclaration without type nor initExpression\n"); }
  ;

varName:  NAMESTRING  { $$=$1; fprintf(yyout, "VarName is %s\n", $1); }
  ;

type: name { $$=$1; fprintf(yyout, "type is \n"); }
  ;

initExpression: OclExpression  { $$=$1; fprintf(yyout, "initExpression\n"); }
  ;


  /* define ModelPropertyCallExp */
ModelPropertyCallExp: OperationCallExp  { $$=A_ModelPropertyCallExp($1,OPERATIONCALLEXP);
                                          fprintf(yyout, "OperationCallExp\n"); }
  | AttrOrNavCallExp  { $$=A_ModelPropertyCallExp($1,ATTRORNAVCALLEXP);
                        fprintf(yyout, "AttributeCallExp or NavigationCallExp\n"); }
  ;

  /* define OperationCallExp
   * source and arguments are replaced by OclExpression, then conflicts can be solved by
   * precedence
   */
OperationCallExp: OclExpression PLUS OclExpression  { $$=A_OperationCallExp($1,"PLUS",FALSE,A_Arguments($3,NULL));
                                                      fprintf(yyout, "PLUS operation\n"); }
  | OclExpression MINUS OclExpression  { $$=A_OperationCallExp($1,"MINUS",FALSE,A_Arguments($3,NULL));
                                         fprintf(yyout, "MINUS operation\n"); }
  | OclExpression TIMES OclExpression  { $$=A_OperationCallExp($1,"TIMES",FALSE,A_Arguments($3,NULL));
                                         fprintf(yyout, "TIMES operation\n"); }
  | OclExpression DIVIDES OclExpression  { $$=A_OperationCallExp($1,"DIVIDES",FALSE,A_Arguments($3,NULL));
                                           fprintf(yyout, "DIVIDES operation\n"); }
  | OclExpression AND OclExpression  { $$=A_OperationCallExp($1,"AND",FALSE,A_Arguments($3,NULL));
                                       fprintf(yyout, "AND operation\n"); }
  | OclExpression OR OclExpression  { $$=A_OperationCallExp($1,"OR",FALSE,A_Arguments($3,NULL));
                                      fprintf(yyout, "OR operation\n"); }
  | OclExpression XOR OclExpression  { $$=A_OperationCallExp($1,"XOR",FALSE,A_Arguments($3,NULL));
                                       fprintf(yyout, "XOR operation\n"); }
  | OclExpression EQ OclExpression  { $$=A_OperationCallExp($1,"EQ",FALSE,A_Arguments($3,NULL));
                                      fprintf(yyout, "EQ operation\n"); }
  | OclExpression NOTEQ OclExpression { $$=A_OperationCallExp($1,"NOTEQ",FALSE,A_Arguments($3,NULL));
                                        fprintf(yyout, "NOTEQ operation\n"); }
  | OclExpression LTEQ OclExpression  { $$=A_OperationCallExp($1,"LTEQ",FALSE,A_Arguments($3,NULL));
                                        fprintf(yyout, "MLTEQ operation\n"); }
  | OclExpression GTEQ OclExpression  { $$=A_OperationCallExp($1,"GTEQ",FALSE,A_Arguments($3,NULL));
                                        fprintf(yyout, "GTEQ operation\n"); }
  | OclExpression LT OclExpression  { $$=A_OperationCallExp($1,"LT",FALSE,A_Arguments($3,NULL));
                                      fprintf(yyout, "LT operation\n"); }
  | OclExpression GT OclExpression  { $$=A_OperationCallExp($1,"GT",FALSE,A_Arguments($3,NULL));
                                      fprintf(yyout, "GT operation\n"); }
  | OclExpression IMPLIES OclExpression  { $$=A_OperationCallExp($1,"IMPLIES",FALSE,A_Arguments($3,NULL));
                                           fprintf(yyout, "IMPLIES operation\n"); }
  | OclExpression DOT mdlProperty timeExpr '(' argumentsOrEmpty ')'
      { $$=A_OperationCallExp($1,$3,$4,$6);
        fprintf(yyout, "model operation\n"); }
  | mdlProperty timeExpr '(' argumentsOrEmpty ')'  { /*$$=A_OperationCallExp(NULL,$1,$2,$4);
						       fprintf(yyout, "model operation with default source\n");*/
                                                     warning("The source of the mdlPropertyCall must be specified\n",NULL); exit(1);}
  | NOT OclExpression  { $$=A_OperationCallExp($2,"NOT",FALSE,NULL);
                         fprintf(yyout, "negates operation\n"); }
  ;

timeExpr:  /* empty */  { $$=FALSE; }
  | ATPRE  { $$=TRUE; fprintf(yyout, "at Pre\n"); }
  ;

argumentsOrEmpty:  /* empty */  { $$=NULL; fprintf(yyout, "empty argument\n"); }
  | arguments { $$=$1; fprintf(yyout, "non empty argumentsn\n"); }
  ;


argument: OclExpression  { $$=$1; fprintf(yyout, "argument\n"); }
  ;

arguments: argument { $$=A_Arguments($1, NULL); }
  | argument ',' arguments  { $$=A_Arguments($1, $3); }
  ;


body: OclExpression { $$=$1; fprintf(yyout, "body\n"); }
  ;

mdlProperty:  NAMESTRING  { $$=$1; fprintf(yyout, "mdlProperty: %s\n", $1); }
  ;

  /* define AttrOrNavCallExp
   * we can only know if it is an AttributeCallExp or a NavigationCallExp
   * when we know the UML model, so this will be done in the next pass
   * source must be specified to resolve the reduce/reduce conflicts
   * thus, "AttrOrNavCallExp::=mdlProperty timeExpr" is not valid
   */
AttrOrNavCallExp: OclExpression DOT mdlProperty timeExpr
    { $$=A_AttrOrNavCallExp($1,$3,$4); fprintf(yyout, "AttrOrNavCallExp with source\n"); }
  ;


  /* define LetExp
   * LetExp will not produce an instance of abstract syntax
   * but it will modify the env
   */
LetExp: LET variables LetExpSub { $$=A_LetExp($2,$3); }
  ;

variables:  variable ';' { $$=A_Variables($1,NULL); }
  | variable ';' variables  { $$=A_Variables($1,$3); fprintf(yyout, "multi-var def\n"); }
  ;

variable: variableDeclarationInLetExp { $$=$1; }
  ;

variableDeclarationInLetExp: varName ':' type EQ initExpression
    { $$=A_VariableDeclaration($1,$3,$5);
      fprintf(yyout, "variableDeclarationInLetExp with type and initExpression\n"); }
  ;

LetExpSub:  IN scope  { $$=$2; fprintf(yyout, "LetExpSub\n"); }
  ;

  /* parenthesis is added to resolve conflicts
   * but it's not the same as the grammar given in the specification
   */
scope:  '(' OclExpression ')'  { $$=$2; fprintf(yyout, "scope of the let expr\n"); }
  ;


  /* define OclMessageExp
   * It need to be studied further
   */
OclMessageExp:  OclExpression '^' NAMESTRING '(' OclMsgArgOrNull ')'
    { $$ = A_OclMessageExp($1, $3, $5); }
  ;

OclMsgArgOrNull:  /* empty */ { $$=NULL; fprintf(yyout, "no args\n"); }
  | OclMessageArgs { $$=$1; fprintf(yyout, "OclMsgArg\n"); }
  ;

OclMessageArgs:  OclMessageArg { $$=$1; }
  | OclMessageArg ',' OclMessageArgs { $$=A_OclMessageArg_link($1,$3); }
  ;

OclMessageArg:  '?' { $$=A_OclMessageArg(A_UnspecifiedValueExp(NULL),UNSPECIFIEDVALUEEXP); }
  | '?' ':' type  { $$=A_OclMessageArg(A_UnspecifiedValueExp($3),UNSPECIFIEDVALUEEXP); }
  | OclExpression  { $$=A_OclMessageArg($1,OCLEXPRESSION); }
  ;


  /* define IfExp
   */
IfExp:  IF condition THEN thenExpression ELSE elseExpression ENDIF
    { $$ = A_IfExp($2, $4, $6); }
  ;

condition: OclExpression { $$ = $1; }
  ;

thenExpression: OclExpression { $$ = $1; }
  ;

elseExpression: OclExpression { $$ = $1; }
  ;

/**************************************************************************************************************
 * for temperorily test
 *************************************************************************************************************/


%%
string progname = "ocl";

#define DEFAULT_TMPCHECK "screen.out"
#define DEFAULT_OUT_NAME "AST_tmp.py"

string default_out_path = "./";

string usage = "usage: %s [oclSource] [tmpCheck]\n";

FILE *AST_out;


int
main(int argc, char **argv)
{
  string oclSource;
  string tmpCheck = NULL;
  char path[255];

  int opt;

  string out_name = DEFAULT_OUT_NAME;
  string function_name = (string)strdup(DEFAULT_OUT_NAME);
  strcpy(path, default_out_path);
  strcat(path, out_name);

  progname = argv[0];

  /* To process command line options */
  while ((opt = getopt(argc, argv, "hi:c:o:")) != EOF)
    {
      switch (opt)
	{
	case 'h':
	  /* print out the usage message */
	  break;
	case 'i':
	  oclSource = optarg;
	  yyin = fopen(oclSource, "r");
	  assert(yyin);
	  break;
	case 'c':
	  tmpCheck = optarg;
	  yyout = fopen (tmpCheck, "w");
	  assert(yyout);
	  break;
	case 'o':
	  strcpy(path, optarg);
	  function_name = strdup(optarg);
	  break;
	default:
	  /* some illegal option, print out the usage info */
	  exit(1);
	}
    }

  if (tmpCheck == NULL)
    {
      tmpCheck = strdup(DEFAULT_TMPCHECK);
      yyout = fopen (tmpCheck, "w");
      assert(yyout);
    }


  /* start to creat python file for the AST in ATOM3 */

  /* remove ".py" from the function name */
  function_name[strlen(function_name) - 3] = '\0';

  AST_out = fopen(path, "w");
  assert(AST_out);

  /* print the header for all the neccessary import files */
  creat_imports();


  /* creat the initial environment */
  creat_init(function_name);

  /* start to parse the Ocl sourse file */
  yyparse();

  creat_end(function_name);

  /* done */
  fclose(AST_out);

  exit(0);
}

void
warning(string s, string t) /* print warning message */
{
  fprintf(stderr, "%s: %s", progname, s);
  if (t)
    fprintf(stderr, " %s", t);
  fprintf(stderr, " line %d\n", lineno);
}













































































